from art.attacks.evasion import *
from .array_utils import *


class TrojEpsAttack:
    def __init__(
        self,
        model,
        learning_rate=0.01,
        eps_steps=0.05,
        max_halving=10,
        max_doubling=10,
        num_iters=15,
        batch_size=128,
        norm=np.inf,
        method = 'minimum',
        k=5,
        use_all=False,
        use_model_preds=False
    ):
        '''

        :param model: Troj/ART classifier instance.
        :param learning_rate: The learning rate for CW L^inf. Only relevant if use_all = True
        :param eps_steps: Size of steps in FGSM and PGD.
        :param max_halving: The max_halving for CW L^inf. Only relevant if use_all = True
        :param max_doubling: The max_doubling for CW L^inf. Only relevant if use_all = True
        :param num_iters: Number of iterations for each attack.
        :param batch_size: Batch size for each attack.
        :param norm: The attack norm for PGD and FGSM.
        :param method: Whether to use the faster loss based method, or the slower minimum method. Argument is given as
        one of 'fast' or 'minimum'
        :param use_model_preds: if minimum is true, whether or not to use the true labels or the model preds for
        evauluating the minimum perturbation
        :param k: When using fast top-k method, this is the size of k.
        :param use_all: Whether to use all 3 basic attacks. Slows down computation, but results in a somewhat stronger
        evaluation.
        '''
        self.model = model
        self.budgets = np.arange(0.1, 1, 0.05)
        self.learning_rate = learning_rate
        self.eps_steps = eps_steps
        self.max_halving = max_halving
        self.max_doubling = max_doubling
        self.num_iters = num_iters
        self.batch_size = batch_size
        self.norm = norm
        self.method = method
        self.k = k
        self.use_all = use_all
        self.use_model_preds = use_model_preds

        #store instantiate attack instances in dictionary
        attack_dict = {}
        for i in range(self.budgets.shape[0]):
            cw_name = "cw_{}".format(i)
            fgm_name = "fgsm_{}".format(i)
            pgd_name = "pgd_{}".format(i)

            attack_dict[fgm_name] = FastGradientMethod(
                self.model,
                norm=self.norm,
                eps=self.budgets[i],
                eps_step=self.eps_steps,
                batch_size=self.batch_size,
            )
            if self.use_all:
                attack_dict[cw_name] = CarliniLInfMethod(self.model,
                                            learning_rate=self.learning_rate, max_iter=self.num_iters,
                                            max_halving=self.max_halving, max_doubling=self.max_doubling,
                                            eps=self.budgets[i], batch_size=self.batch_size, verbose=False)

                attack_dict[pgd_name] = ProjectedGradientDescent(self.model, norm=self.norm, eps=self.budgets[i],
                                                       eps_step=self.eps_steps,
                                                       batch_size=self.batch_size, max_iter=self.num_iters, verbose=False)
        self.attack_dict = attack_dict

    def generate(self, x, y):
        """
        :param x: inputs
        :param y: labels, either true labels or original unperturbed model labels. y might need to be expanded along
        the first dimension because of art bug.
        :return: adversarial examples, adversarial losses, adversarial predictions
        """
        generated_examples = []
        model_adv_losses = []
        model_adv_preds = []
        #For each attack method in the attack dictionary, generate adversarial examples, then compute the loss
        #and the class predictions.
        for attacker in list(self.attack_dict.values()):
            adv_x = attacker.generate(x, y)
            if len(y.shape) == 1:
                adv_losses, adv_preds = self.model.ComputeLoss(adv_x, y)
                if self.use_model_preds == True:
                    true_losses, true_preds = self.model.ComputeLoss(x, y)
            else:
                adv_losses, adv_preds = self.model.ComputeLoss(adv_x, np.squeeze(y, axis=1))
                if self.use_model_preds == True:
                    true_losses, true_preds = self.model.ComputeLoss(x, np.squeeze(y, axis=1))
            model_adv_losses.append(adv_losses)
            generated_examples.append(adv_x)
            model_adv_preds.append(adv_preds)
        #reshape arrays so that they each have shape [batch_size, num_attacks, *]
        generated_examples = np.stack(generated_examples)
        generated_examples = np.swapaxes(generated_examples, 0, 1)

        model_adv_losses = np.stack(model_adv_losses)
        model_adv_losses = np.swapaxes(model_adv_losses, 0, 1)

        model_adv_preds = np.stack(model_adv_preds)
        model_adv_preds = np.swapaxes(model_adv_preds, 0, 1)

        #Whether or not to compute the perturbations quickly, or in such a way as to minimize
        # the perturbation that flips the label.
        if self.method=='fast':
            high_loss_examples, losses, sorted_preds = AdvLossSort(generated_examples, model_adv_losses, model_adv_preds, k=self.k)
            output, preds, losses = get_min_pert(x, high_loss_examples, losses, sorted_preds, norm=self.norm)
        elif self.method=='minimum':
            if self.use_model_preds==True:
                labels = true_preds
            else:
                labels = y
            output, preds, losses = true_min_pert(labels, model_adv_preds, x, generated_examples, model_adv_losses)

        return output, preds, losses



from . import TrojEpsilon
import numpy as np
# import pandas as pd
from . import array_utils
from . import attack_utils
from . import ODAttack
import torch


class BasicRobustnessEvaluator:
    def __init__(
        self,
        classifier,
        learning_rate=0.01,
        eps_steps=0.05,
        max_halving=10,
        max_doubling=10,
        num_iters=15,
        batch_size=128,
        norm=np.inf,
        method="minimum",
        k=5,
        use_all=False,
        use_model_preds=False,
    ):
        self.classifier = classifier
        self.attacker = TrojEpsilon.TrojEpsAttack(
            self.classifier,
            learning_rate=learning_rate,
            eps_steps=eps_steps,
            max_halving=max_halving,
            max_doubling=max_doubling,
            num_iters=num_iters,
            batch_size=batch_size,
            norm=norm,
            method=method,
            k=k,
            use_all=use_all,
            use_model_preds=use_model_preds,
        )

    def attack(self, data, target, index, device=None):
        """
        This will work for pytorch, tensorflow has no device
        """
        # index = index.numpy()
        # send data and target to cpu, convert to numpy
        data = np.ascontiguousarray(data.astype(np.float32))
        test_loss, preds = self.classifier.ComputeLoss(data, target)
        preds = np.argmax(preds, axis=1)
        adv_x, adv_preds, adv_loss = self.attacker.generate(data, target)
        # adv_loss, adv_preds = self.classifier.ComputeLoss(adv_x, target)
        perturbation = array_utils.compute_Lp_distance(data, adv_x)
        adv_pred = np.argmax(adv_preds, axis=1)
        # generate the adversarial image using the data numpy array and label numpy array
        out_dict = {
            "Linf_perts": perturbation,
            "Loss": test_loss,
            "Adversarial_Loss": adv_loss,
            "prediction": preds,
            "Adversarial_prediction": adv_pred,
        }
        return (out_dict, index)


class BlackBoxODEvaluator:
    # different run form to classification
    def __init__(
        self,
        model,
        obj_class,
        loader,
        batch_iterator,
        df=None,
        device="cuda",
        iou_thresh=0.5,
        nms_thresh=0.05,
        step_size=0.1,
        verbose=True,
        return_prc=True,
        **attkwargs
    ):
        self.model = model
        self.obj_class = obj_class
        self.loader = loader
        self.batch_iterator = batch_iterator
        self.df = df
        if self.df == None:
            self.df = loader.dataframe
        self.device = device
        self.iou_thresh = iou_thresh
        self.nms_thresh = nms_thresh
        self.step_size = step_size
        self.attacker = ODAttack.EvoDAttack(self.model, self.obj_class, **attkwargs)
        self.verbose = verbose
        self.return_prc = return_prc

    def run(self, num_samples):
        # TODO: Should the dataset task checker be in here?
        tracker = 0
        attacked_ids = []
        batch_enum = enumerate(self.batch_iterator)

        while tracker < num_samples:
            batch_id, (ims, labs, ids) = next(batch_enum)

            if tracker > 0 and batch_id == 0:
                break

            for idx in range(len(labs)):
                sample_id = ids[idx]
                data_dict = {}
                perturb, gt, preds = self.attacker.attack(ims[idx], labs[idx])
                if gt != None:
                    pert_im = ims[idx] + perturb.to(self.device)
                    pert_preds = self.model([pert_im])[0]
                    nms_pert_preds = ODAttack.nms_pred_reduce(pert_preds, self.nms_thresh)
                    nms_preds = ODAttack.nms_pred_reduce(preds, self.nms_thresh)
                    flip = ODAttack.check_flip(nms_pert_preds, gt, self.obj_class, self.iou_thresh, self.nms_thresh)

                    if self.return_prc:
                        troj_map, perc, rec, thresh = ODAttack.mAP(nms_preds, labs[idx], self.step_size, self.iou_thresh)
                        adv_troj_map, adv_perc, adv_rec, adv_thresh = ODAttack.mAP(nms_pert_preds, labs[idx], self.step_size, self.iou_thresh)
                    else:
                        troj_map = ODAttack.mAP(nms_preds, labs[idx], self.step_size, self.iou_thresh)
                        adv_troj_map = ODAttack.mAP(nms_pert_preds, labs[idx], self.step_size, self.iou_thresh)
                        
                    pert_vec = perturb.view(-1)
                    linf_pert = torch.norm(pert_vec, p=np.inf)
                    tracker += 1

                    if self.return_prc:
                        data_dict = {
                            "flip": flip,
                            "TmAP": troj_map,
                            "Adv_TmAP": adv_troj_map,
                            "Linf": linf_pert.item(),
                            "precision": perc,
                            "recall": rec,
                            "thresh": thresh,
                            "adv_precision": adv_perc,
                            "adv_recall": adv_rec,
                            "adv_thresh": adv_thresh,
                        }
                    else:
                        data_dict = {
                            "flip": flip,
                            "TmAP": troj_map,
                            "Adv_TmAP": adv_troj_map,
                            "Linf": linf_pert.item(),
                        }

                    self.df = attack_utils.log_to_dataframe(self.df, sample_id, data_dict)
                    attacked_ids.append(sample_id)

        # Insert the model used into the df <- not going to work..
        # Inserting into DB would be nice, how to get dataset_uuid/max_metric_job_uuid ?
        if "model_used" not in self.df:
            self.df["model_used"] = ""
            self.df["model_used"].astype("object")
        # dataframe.loc[index, key] = log_dict[key]
        self.df["model_used"] = self.model.__class__.__name__
        return self.df, attacked_ids



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="troj",
    version="0.1.0.1",
    packages=["trojai"],
    author="TrojAI",
    author_email="stan.petley@troj.ai",
    description="TrojAI provides the troj Python convenience package to allow users to integrate TrojAI adversarial protections and robustness metrics seamlessly into their AI development pipelines.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://troj.ai",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "torch>=1.6.0",
        "tensorflow>=2.0.0",
        "requests",
        "urllib3",
        "adversarial-robustness-toolbox==1.5.2",
        "opencv-python",
    ],
    python_requires=">=3.6",
    extras_require={
        "dev": [
            "pytest>=6.2",
        ],
    },
    # Dependencies go here
    # install_requires = [
    #     "library ~= N.M",
    # ]
)
